@@ -54,6 +54,7 @@ class _State:
         self.T, self.complete, self.items = [], [], items
         self.stateno = stateno
 
+
 # DEFAULT_DEBUG = {'rules': True, 'transition': True, 'reduce' : True,
 #                  'errorstack': 'full', 'dups': False }
 # DEFAULT_DEBUG = {'rules': False, 'transition': False, 'reduce' : True,
@@ -61,6 +62,7 @@ class _State:
 DEFAULT_DEBUG = {'rules': False, 'transition': False, 'reduce': False,
                  'errorstack': None, 'context': True, 'dups': False}
 
+
 class GenericParser(object):
     '''
     An Earley parser, as per J. Earley, "An Efficient Context-Free
@@ -98,6 +100,15 @@ class GenericParser(object):
         # When set, shows additional debug output
         self.debug = debug
 
+        # Have a place to tag list-like rules. These include rules of the form:
+        #   a ::= x+
+        #   b ::= x*
+        #
+        # These kinds of rules, we should create as a list when building a
+        # parse tree rather than a sequence of nested derivations
+        self.list_like_nt = set()
+        self.optional_nt = set()
+
         self.collectRules()
         if start not in self.rules:
             raise TypeError('Start symbol "%s" is not in LHS of any rule' % start)
@@ -211,12 +222,15 @@ class GenericParser(object):
                 if rule[1][0] == rule[0]:
                     raise TypeError("Complete recursive rule %s" % rule2str(rule))
 
-                if rule[1][-1][-1] in ('*', '+', '?'):
-                    repeat = rule[1][-1][-1]
+                repeat = rule[1][-1][-1]
+                if repeat in ('*', '+', '?'):
                     nt = rule[1][-1][:-1]
+
                     if repeat == '?':
                         new_rule_pair = [rule[0], list((nt,))]
+                        self.optional_nt.add(rule[0])
                     else:
+                        self.list_like_nt.add(rule[0])
                         new_rule_pair = [rule[0], [rule[0]] + list((nt,))]
                     new_rule = rule2str(new_rule_pair)
                     self.addRule(new_rule, func, _preprocess)
@@ -240,7 +254,10 @@ class GenericParser(object):
             self.rule2name[rule] = func.__name__[2:]
             self.ruleschanged = True
 
-            if self.profile_info is not None:
+            # Note: In empty rules, i.e. len(rule[1] == 0, we don't
+            # call reductions on explicitly. Instead it is computed
+            # implicitly.
+            if self.profile_info is not None and len(rule[1]) > 0:
                 rule_str = self.reduce_string(rule)
                 if rule_str not in self.profile_info:
                     self.profile_info[rule_str] = 0
@@ -251,6 +268,9 @@ class GenericParser(object):
         """Remove a grammar rules from  _self.rules_, _self.rule2func_,
             and _self.rule2name_
         """
+        # remove blanks lines and comment lines, e.g. lines starting with "#"
+        doc = os.linesep.join([s for s in doc.splitlines() if s and not re.match("^\s*#", s)])
+
         rules = doc.split()
         index = []
         for i in range(len(rules)):
@@ -270,6 +290,16 @@ class GenericParser(object):
                 del self.rule2func[rule]
                 del self.rule2name[rule]
                 self.ruleschanged = True
+
+                # If we are profiling, remove this rule from that as well
+                if self.profile_info is not None and len(rule[1]) > 0:
+                    rule_str = self.reduce_string(rule)
+                    if rule_str and rule_str in self.profile_info:
+                        del self.profile_info[rule_str]
+                        pass
+                    pass
+                pass
+
         return
 
     remove_rule = remove_rules
@@ -283,14 +313,21 @@ class GenericParser(object):
 
     def augment(self, start):
         rule = '%s ::= %s %s' % (self._START, self._BOF, start)
-        self.addRule(rule, lambda args: args[1], 0)
+        self.addRule(rule, lambda args: args[1], False)
 
     def computeNull(self):
         self.nullable = {}
         tbd = []
 
         for rulelist in list(self.rules.values()):
+
+            # FIXME: deleting a rule may leave a null entry.
+            # Perhaps we should improve deletion so it doesn't leave a trace?
+            if not rulelist:
+                continue
+
             lhs = rulelist[0][0]
+
             self.nullable[lhs] = 0
             for rule in rulelist:
                 rhs = rule[1]
@@ -348,9 +385,9 @@ class GenericParser(object):
             while i < n:
                 sym = rhs[i]
                 if (sym not in self.rules or
-                    not self.nullable[sym]):
+                    not (sym in self.nullable and self.nullable[sym])):
                         candidate = 0
-                        i = i + 1
+                        i += 1
                         continue
 
                 newrhs = list(rhs)
@@ -405,9 +442,16 @@ class GenericParser(object):
                 lhs, rhs = rule
                 if dot > 0:
                     if full:
-                        state_stack.add(' '.join(rhs[:dot]) + ' . ' + ' '.join(rhs[dot:]))
+                        state_stack.add(
+                            "%s ::= %s . %s" %
+                            (lhs,
+                            ' '.join(rhs[:dot]),
+                            ' '.join(rhs[dot:])))
                     else:
-                        state_stack.add(' '.join(rhs[:dot]))
+                        state_stack.add(
+                            "%s ::= %s" %
+                            (lhs,
+                            ' '.join(rhs[:dot])))
                     pass
                 pass
             pass
@@ -421,6 +465,8 @@ class GenericParser(object):
         setting.
         """
 
+        self.tokens = tokens
+
         if debug:
             self.debug = debug
 
@@ -450,7 +496,7 @@ class GenericParser(object):
         finalitem = (self.finalState(tokens), 0)
         if finalitem not in sets[-2]:
             if len(tokens) > 0:
-                if self.debug['errorstack']:
+                if self.debug.get('errorstack', False):
                     self.errorstack(tokens, i-1, str(self.debug['errorstack']) == 'full')
                 self.error(tokens, i-1)
             else:
@@ -629,12 +675,15 @@ class GenericParser(object):
                     self.debug_reduce(rule, tokens, parent, i)
                 if self.profile_info is not None:
                     self.profile_rule(rule)
-                if lhs in self.check_reduce and tokens:
-                    if self.check_reduce[lhs] == 'AST':
-                        ast = self.reduce_ast(rule, tokens, item, i, sets)
+                if lhs in self.check_reduce:
+                    if (self.check_reduce[lhs] == 'AST' and
+                        (tokens or hasattr(self, 'tokens'))):
+                        if hasattr(self, 'tokens'):
+                            tokens = self.tokens
+                        ast = self.reduce_ast(rule, self.tokens, item, i, sets)
                     else:
                         ast = None
-                    invalid = self.reduce_is_invalid(rule, ast, tokens, parent, i)
+                    invalid = self.reduce_is_invalid(rule, ast, self.tokens, parent, i)
                     if ast:
                         del ast
                     if invalid:
@@ -642,6 +691,7 @@ class GenericParser(object):
                             print("Reduce %s invalid by check" % lhs)
                         continue
                     pass
+                    pass
                 for pitem in sets[parent]:
                     pstate, pparent = pitem
                     k = self.goto(pstate, lhs)
@@ -821,7 +871,7 @@ class GenericParser(object):
         '''
         return list[0]
 
-    def dumpGrammar(self, out=sys.stdout):
+    def dump_grammar(self, out=sys.stdout):
         """
         Print grammar rules
         """
@@ -829,37 +879,65 @@ class GenericParser(object):
             out.write("%s\n" % rule2str(rule[0]))
         return
 
-    def checkGrammar(self, out=sys.stderr):
+    def check_grammar(self, ok_start_symbols = set(),
+                      out=sys.stderr):
         '''
-        Check grammar
+        Check grammar for:
+        -  unused left-hand side nonterminals that are neither start symbols
+           or listed in ok_start_symbols
+        -  unused right-hand side nonterminals, i.e. not tokens
+        -  right-recursive rules. These can slow down parsing.
         '''
-        lhs, rhs, tokens, right_recursive = self.checkSets()
-        if len(lhs) > 0:
+        warnings = 0
+        (lhs, rhs, tokens, right_recursive,
+         dup_rhs) = self.check_sets()
+        if lhs - ok_start_symbols:
+            warnings += 1
             out.write("LHS symbols not used on the RHS:\n")
-            out.write(sorted(lhs), "\n")
-        if len(rhs) > 0:
+            out.write("  " + (', '.join(sorted(lhs)) + "\n"))
+        if rhs:
+            warnings += 1
             out.write("RHS symbols not used on the LHS:\n")
-            out.write(sorted(rhs, "\n"))
-        if len(right_recursive) > 0:
+            out.write((', '.join(sorted(rhs))) + "\n" )
+        if right_recursive:
+            warnings += 1
             out.write("Right recursive rules:\n")
-            for rule in right_recursive:
-                out.write("%s ::= %s\n" % (rule[0], ' '.join(rule[1])))
+            for rule in sorted(right_recursive):
+                out.write("  %s ::= %s\n" % (rule[0], ' '.join(rule[1])))
                 pass
             pass
+        if dup_rhs:
+            warnings += 1
+            out.write("Nonterminals with the same RHS\n")
+            for rhs in sorted(dup_rhs.keys()):
+                out.write("  RHS: %s\n" % ' '.join(rhs))
+                out.write("  LHS: %s\n" % ', '.join(dup_rhs[rhs]))
+                out.write("  ---\n")
+                pass
+            pass
+        return warnings
 
-    def checkSets(self):
+    def check_sets(self):
         '''
         Check grammar
         '''
-        lhs_set = set()
-        rhs_set = set()
-        token_set = set()
-        right_recursive = []
+        lhs_set         = set()
+        rhs_set         = set()
+        rhs_rules_set   = {}
+        token_set       = set()
+        right_recursive = set()
+        dup_rhs = {}
         for lhs in self.rules:
             rules_for_lhs = self.rules[lhs]
             lhs_set.add(lhs)
             for rule in rules_for_lhs:
                 rhs = rule[1]
+                if len(rhs) > 0 and rhs in rhs_rules_set:
+                    li = dup_rhs.get(rhs, [])
+                    li.append(lhs)
+                    dup_rhs[rhs] = li
+                else:
+                    rhs_rules_set[rhs] = lhs
                 for sym in rhs:
                     # We assume any symbol starting with an uppercase letter is
                     # terminal, and anything else is a nonterminal
@@ -868,7 +946,7 @@ class GenericParser(object):
                     else:
                         rhs_set.add(sym)
                 if len(rhs) > 0 and lhs == rhs[-1]:
-                    right_recursive.append([lhs, rhs])
+                    right_recursive.add((lhs, rhs))
                 pass
             pass
 
@@ -876,15 +954,23 @@ class GenericParser(object):
         rhs_set.remove(self._BOF)
         missing_lhs = lhs_set - rhs_set
         missing_rhs = rhs_set - lhs_set
-        return (missing_lhs, missing_rhs, token_set, right_recursive)
 
-    def reduce_string(self, rule):
-        return "%s ::= %s" % (rule[0], ' '.join(rule[1]))
+        # dup_rhs is missing first entry found, so add that
+        for rhs in dup_rhs:
+            dup_rhs[rhs].append(rhs_rules_set[rhs])
+            pass
+        return (missing_lhs, missing_rhs, token_set, right_recursive, dup_rhs)
+
+    def reduce_string(self, rule, last_token_pos=-1):
+        if last_token_pos >= 0:
+            return "%s ::= %s (%d)" % (rule[0], ' '.join(rule[1]), last_token_pos)
+        else:
+            return "%s ::= %s" % (rule[0], ' '.join(rule[1]))
 
     # Note the unused parameters here are used in subclassed
     # routines that need more information
     def debug_reduce(self, rule, tokens, parent, i):
-        print(self.reduce_string(rule))
+        print(self.reduce_string(rule, i))
 
     def profile_rule(self, rule):
         """Bump count of the number of times _rule_ was used"""
